/*
 * @(#)FileDescriptor.java	1.22 05/11/17
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package java.io;

/**
 * Instances of the file descriptor class serve as an opaque handle
 * to the underlying machine-specific structure representing an open
 * file, an open socket, or another source or sink of bytes. The
 * main practical use for a file descriptor is to create a
 * <code>FileInputStream</code> or <code>FileOutputStream</code> to
 * contain it.
 * <p>
 * Applications should not create their own file descriptors.
 *
 * @author  Pavani Diwanji
 * @version 1.22, 11/17/05
 * @see	    java.io.FileInputStream
 * @see	    java.io.FileOutputStream
 * @since   JDK1.0
 */
/*                                                                              //IBM-perf_FileIO
 *  Date      Work-Item  Author           Comment                               //IBM-perf_FileIO
 *  ========  =========  ===============  =================================     //IBM-perf_FileIO
 *  20061206  JSE-771    Mark Bluemel     Performance enhancements from         //IBM-perf_FileIO
 *                                        Sovereign Feature 68577               //IBM-perf_FileIO
 */                                                                             //IBM-perf_FileIO
public final class FileDescriptor {

    private int fd;

    /**
     * Reference Count, to ensure that shared FileDescriptors                   //IBM-FileDescriptor_reference_counting
     * are closed by the last object to be using them                           //IBM-FileDescriptor_reference_counting
     */                                                                         //IBM-FileDescriptor_reference_counting
    private int refCount = 0;                                                   //IBM-FileDescriptor_reference_counting
                                                                                //IBM-FileDescriptor_reference_counting
    /**                                                                         //IBM-FileDescriptor_reference_counting
     * Constructs an (invalid) FileDescriptor
     * object.
     */
    public /**/ FileDescriptor() {
	fd = -1;
    }

    private /* */ FileDescriptor(int fd) {
	this.fd = fd;
    }

    int read() throws IOException {                                             //IBM-perf_FileIO
        return read(fd);                                                        //IBM-perf_FileIO
    }                                                                           //IBM-perf_FileIO
                                                                                //IBM-perf_FileIO
    private native int read(long fd) throws IOException;                        //IBM-perf_FileIO
                                                                                //IBM-perf_FileIO
                                                                                //IBM-perf_FileIO
    /** JSE-4336: Added native methods to fetch the respective                  //IBM-perf_FileIO
     *  FileDescriptors for STDIN,STDOUT and STDERR since the                   //IBM-perf_FileIO
     *  default values of 0,1 and 2 fail for non-shell environment              //IBM-perf_FileIO
     *  in zOS                                                                  //IBM-perf_FileIO
     */	                                                                 //IBM-perf_FileIO
    private static native int getStdinFD();                                      //IBM-perf_FileIO
    private static native int getStdoutFD();                                     //IBM-perf_FileIO
    private static native int getStderrFD();                                     //IBM-perf_FileIO
                                                                                 //IBM-perf_FileIO
                                                                                //IBM-perf_FileIO
    /**
     * A handle to the standard input stream. Usually, this file
     * descriptor is not used directly, but rather via the input stream
     * known as <code>System.in</code>.
     *
     * @see     java.lang.System#in
     */
    public static final FileDescriptor in = new FileDescriptor(getStdinFD());   //JSE-4336 //IBM-perf_FileIO

    /**
     * A handle to the standard output stream. Usually, this file
     * descriptor is not used directly, but rather via the output stream
     * known as <code>System.out</code>.
     * @see     java.lang.System#out
     */
    public static final FileDescriptor out = new FileDescriptor(getStdoutFD()); //JSE-4336 //IBM-perf_FileIO

    /**
     * A handle to the standard error stream. Usually, this file
     * descriptor is not used directly, but rather via the output stream
     * known as <code>System.err</code>.
     *
     * @see     java.lang.System#err
     */
    public static final FileDescriptor err = new FileDescriptor(getStderrFD()); //JSE-4336 //IBM-perf_FileIO

    /**
     * Tests if this file descriptor object is valid.
     *
     * @return  <code>true</code> if the file descriptor object represents a
     *          valid, open file, socket, or other active I/O connection;
     *          <code>false</code> otherwise.
     */
    public boolean valid() {
	return fd != -1;
    }

    /**
     * Force all system buffers to synchronize with the underlying
     * device.  This method returns after all modified data and
     * attributes of this FileDescriptor have been written to the
     * relevant device(s).  In particular, if this FileDescriptor
     * refers to a physical storage medium, such as a file in a file
     * system, sync will not return until all in-memory modified copies
     * of buffers associated with this FileDesecriptor have been
     * written to the physical medium.
     *
     * sync is meant to be used by code that requires physical
     * storage (such as a file) to be in a known state  For
     * example, a class that provided a simple transaction facility
     * might use sync to ensure that all changes to a file caused
     * by a given transaction were recorded on a storage medium.
     *
     * sync only affects buffers downstream of this FileDescriptor.  If
     * any in-memory buffering is being done by the application (for
     * example, by a BufferedOutputStream object), those buffers must
     * be flushed into the FileDescriptor (for example, by invoking
     * OutputStream.flush) before that data will be affected by sync.
     *
     * @exception SyncFailedException
     *	      Thrown when the buffers cannot be flushed,
     *	      or because the system cannot guarantee that all the
     *	      buffers have been synchronized with physical media.
     * @since     JDK1.1
     */
    public native void sync() throws SyncFailedException;

   /**                                                                          //IBM-FileDescriptor_reference_counting
    * Increment the reference count to reflect the fact                         //IBM-FileDescriptor_reference_counting
    * that another object is using this FileDescriptor                          //IBM-FileDescriptor_reference_counting
    */                                                                          //IBM-FileDescriptor_reference_counting
   synchronized void incRefCount() {                                            //IBM-FileDescriptor_reference_counting
        refCount++;                                                             //IBM-FileDescriptor_reference_counting
    }                                                                           //IBM-FileDescriptor_reference_counting
                                                                                //IBM-FileDescriptor_reference_counting
    /**                                                                         //IBM-FileDescriptor_reference_counting
     * Decrement the reference count to note that an                            //IBM-FileDescriptor_reference_counting
     * object is no longer using this FileDescriptor                            //IBM-FileDescriptor_reference_counting
     * @return the reference count - when 0, the decementing object             //IBM-FileDescriptor_reference_counting
     *         can safely close() the FileDescriptor                            //IBM-FileDescriptor_reference_counting
     */                                                                         //IBM-FileDescriptor_reference_counting
    synchronized int decRefCount() {                                            //IBM-FileDescriptor_reference_counting
        refCount--;                                                             //IBM-FileDescriptor_reference_counting
        return refCount;                                                        //IBM-FileDescriptor_reference_counting
    }                                                                           //IBM-FileDescriptor_reference_counting
                                                                                //IBM-FileDescriptor_reference_counting
    /* This routine initializes JNI field offsets for the class */
    private static native void initIDs();

    static {
	initIDs();
    }
}
//IBM-FileDescriptor_reference_counting
//IBM-perf_FileIO
